#!/usr/bin/env python

"""
Merge page packages.

Copyright (C) 2013 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from os import chdir, getcwd, makedirs, mkdir, walk
from os.path import exists, extsep, join, normpath, relpath, split, splitext
from shutil import copy
from zipfile import ZipFile
from cStringIO import StringIO
import sys

def get_filenames(package):
    results = []
    for path, dirnames, filenames in walk(package):
        path = relpath(path, package)
        for filename in filenames:
            results.append(join(path, filename))
    return results

# Main program.

if __name__ == "__main__":
    packages = sys.argv[2:]

    if not packages:
        print >>sys.stderr, """
Please specify an output basename followed by a list of page packages.
For example:

%(progname)s OUT COM.zip DEV.zip DOC.zip SEC.zip

As a result of running this program, a page package will be created at the
specified basename along with an archive of the form OUT.zip given a basename of
OUT.
""" % {"progname" : split(sys.argv[0])[-1]}

        sys.exit(1)

    outdir = sys.argv[1]
    outleafname = split(outdir)[-1]

    if exists(outdir):
        print >>sys.stderr, "Directory %s exists already. Please choose another or remove its contents." % outdir
        sys.exit(1)

    package_zip = outdir + extsep + "zip"

    if exists(package_zip):
        print >>sys.stderr, "Page package exists. Please remove or rename it:", package_zip
        sys.exit(1)

    # Make the output directory for the merged packages.

    mkdir(outdir)

    # Collect entries from all packages for sorting.

    entries = []

    for package in packages:
        is_zipfile = splitext(package)[-1] == extsep + "zip"
        leafname = split(package)[-1]

        if is_zipfile:
            f = open(package, "rb")
            zf = ZipFile(f)
            ff = StringIO(zf.read("MOIN_PACKAGE"))
        else:
            ff = open(join(package, "MOIN_PACKAGE"))

        try:
            # Skip the first line and get the manifest entries.
            # NOTE: We could use the MoinMoin.package API here.

            ff.readline()
            entries += [x.rstrip("\n").split("|") for x in ff.readlines()]

            # Copy files from the package into the output directory.

            if is_zipfile:
                filenames = zf.namelist()
            else:
                filenames = get_filenames(package)

            for filename in filenames:
                if split(filename)[-1] == "MOIN_PACKAGE":
                    continue

                # Extract files, tidying up any filesystem pathnames.

                if is_zipfile:
                    zf.extract(filename, outdir)
                else:
                    target = normpath(join(outdir, leafname, filename))
                    target_dir = split(target)[0]
                    if not exists(target_dir):
                        makedirs(target_dir)
                    copy(join(package, filename), target)

        finally:
            ff.close()
            if is_zipfile:
                zf.close()

    # The final entry in each element is the mtime.

    entries.sort(cmp=lambda x, y: cmp(int(x[-1]), int(y[-1])))

    # Write the combined manifest to the output directory.

    output_manifest = join(outdir, "MOIN_PACKAGE")

    f = open(output_manifest, "w")
    write = f.write
    try:
        write("MoinMoinPackage|1\n")
        for entry in entries:

            # Reference the adjusted location of each file.

            entry[1] = join(outleafname, entry[1])
            write("|".join(entry) + "\n")
    finally:
        f.close()

    # Write the page package.

    page_package = ZipFile(package_zip, "w")

    try:
        for filename in get_filenames(outdir):
            page_package.write(join(outleafname, filename))
    finally:
        page_package.close()

# vim: tabstop=4 expandtab shiftwidth=4
